from .models import MusicFile, Tag, MusicFileTag, Preferences
from django.contrib.auth.models import User
from django.db import connection
from typing import List
from .fts import fts_escape


def sanitize_tag_label(tag_label: str) -> str:
    return " ".join(tag_label.split()).lower().strip()


def queue_list(owner: User):
    return MusicFileTag.objects.filter(tag__owner=owner, tag__label="queue").order_by(
        "rank"
    )


def add_tag_for_owner(label: str, owner: User) -> bool:
    label = sanitize_tag_label(label)
    if label:
        tag_instance, _ = Tag.objects.get_or_create(owner=owner, label=label)
        return True
    return False


def remove_tag_for_owner(tag_pk: int, owner: User) -> bool:
    Tag.objects.filter(pk=tag_pk, owner=owner).exclude(
        label__in=["queue", "favorites"]
    ).delete()


def add_music_file_to_tag(
    music_file: MusicFile, owner: User, tag: str, rank: int = 0
) -> MusicFileTag:
    tag_instance, _ = Tag.objects.get_or_create(
        owner=owner, label=sanitize_tag_label(tag)
    )
    # when adding track to tag, set rank if not given
    if rank == 0:
        last_in_queue = (
            MusicFileTag.objects.filter(tag=tag_instance).order_by("-rank").first()
        )
        if last_in_queue:
            rank = last_in_queue.rank + 1

    assoc, created = MusicFileTag.objects.get_or_create(
        music_file=music_file, tag=tag_instance, defaults={"rank": rank}
    )
    if not created and assoc.rank != rank:
        assoc.rank = rank
        assoc.save()
    return assoc


def remove_music_file_from_tag(music_file: MusicFile, owner: User, tag: str) -> bool:
    tag = sanitize_tag_label(tag)
    MusicFileTag.objects.filter(
        music_file=music_file, tag__owner=owner, tag__label=tag
    ).delete()
    return True


def add_music_file_to_favorites(music_file: MusicFile, owner: User) -> MusicFileTag:
    return add_music_file_to_tag(music_file, owner, "favorites")


def remove_music_file_from_favorites(music_file: MusicFile, owner: User) -> bool:
    return remove_music_file_from_tag(music_file, owner, "favorites")


def music_file_is_favorite(music_file: MusicFile, owner: User) -> bool:
    return MusicFileTag.objects.filter(
        music_file=music_file, tag__owner=owner, tag__label="favorites"
    ).exists()


def user_tags(user: User):
    tags_list = (
        # Tag.objects.filter(owner=user, musicfiletag__isnull=False)
        Tag.objects.filter(owner=user)
        .exclude(label__in=["queue", "favorites"])
        .distinct()
        .order_by("label")
    )
    return tags_list


def distinct_fuzzy(column: str, q: str) -> List[str]:
    distincts = []
    sql = f"""
    select distinct(reinhardt_musicfile.{column})
        from reinhardt_musicfile
        inner join music_index on music_index.id = reinhardt_musicfile.id
        where music_index.{column} match %s
        order by reinhardt_musicfile.{column};
    """
    with connection.cursor() as cursor:
        cursor.execute(sql, (fts_escape(q),))
        records = cursor.fetchall()
        for row in records:
            distincts.append(row[0])
    return distincts


def user_preferences(user: User) -> Preferences:
    prefs, _ = Preferences.objects.get_or_create(owner=user)
    return prefs
