from django.core.management.base import BaseCommand
from reinhardt.models import MusicDirectory, MusicFile
from reinhardt.fts import fts_index_music_files
from typing import Iterator
import os
import tinytag
from datetime import datetime
from django.utils.timezone import make_aware
import logging
from django.core.cache import cache
import re
from django.conf import settings

logger = logging.getLogger(__name__)
RE_PATTERN_MATCH_ID3V1_GENRE = re.compile(r"^\([0-9]+\).+")


def ci_replace(words: str, old: str, new: str) -> str:
    pattern = re.compile(re.escape(old), re.IGNORECASE)
    return pattern.sub(new, words)


def sanitize_tag_value(value: str, tag: str = None) -> str:
    if not settings.SANITIZE_TAGS:
        return value
    #
    # ID3 Tags are a big mess... gotta deal with shitty data.
    #
    # force split of strings with slashes and ampersands
    value = value.replace("/", " / ")
    value = value.replace("&", " & ")
    # remove and deduplicate non-printable chars
    value = " ".join(value.strip().split())
    # "Something" -> Something
    if len(value) > 1 and value[0] == '"' and value[-1] == '"':
        value = value[1:-1]
    if tag == "artist":
        # Heim & Jones -> Heim and Jones
        value = ci_replace(value, " & ", " and ")
        value = ci_replace(value, " feat. ", " featuring ")
        value = ci_replace(value, " feat ", " featuring ")
        value = ci_replace(value, " vs. ", " versus ")
        value = ci_replace(value, " vs ", " versus ")
        # roni size -> Roni Size
        value = " ".join(x.capitalize() for x in value.lower().split())
        # The Beatles -> Beatles (the)
        if len(value) > 4 and value[0:4].lower() == "the ":
            value = f"{value[4:]} (the)"
        # DJ Gizmo -> Gizmo (DJ)
        if len(value) > 3 and value[0:3].lower() == "dj ":
            value = f"{value[3:]} (DJ)"
        if value.lower() in ["various", "various artist", "various artists"]:
            value = "various"
    if tag == "genre":
        # Techno -> techno
        value = value.lower()
        value = ci_replace(value, " & ", " and ")
        # (18)techno -> techno
        if RE_PATTERN_MATCH_ID3V1_GENRE.match(value):
            value = value.split(")")[1]
    return value


def _data_from_tags(path: str, base_dir: str = None) -> dict:
    data = {
        "artist": "",
        "title": "",
        "album": "",
        "genre": "",
        "comment": "",
        "duration": 0,
        "track": 0,
        "relative_path": path[len(base_dir) :].lstrip("/") if base_dir else path,
        "full_path": path,
        "creation": make_aware(datetime.fromtimestamp(os.path.getctime(path))),
        "modification": make_aware(datetime.fromtimestamp(os.path.getmtime(path))),
    }
    tag = tinytag.TinyTag.get(path, ignore_errors=True)
    if tag.artist:
        data["artist"] = sanitize_tag_value(tag.artist, "artist")
    if tag.title:
        data["title"] = sanitize_tag_value(tag.title)
    if tag.album:
        data["album"] = sanitize_tag_value(tag.album)
    if tag.genre:
        data["genre"] = sanitize_tag_value(tag.genre, "genre")
    if tag.comment:
        data["comment"] = tag.comment.strip()
    if tag.duration:
        data["duration"] = tag.duration
    if tag.track:
        try:
            val = tag.track.split("/")[0]
            val = int(val)
            data["track"] = val
        except ValueError:
            pass
    return data


def _walk_music_files(root_dir: str) -> Iterator[str]:
    for f_root, _, all_files in os.walk(root_dir):
        for f_name in all_files:
            if f_name.lower().endswith(
                (
                    ".mp3",
                    ".m4a",
                )
            ):
                yield os.path.join(f_root, f_name)


class Command(BaseCommand):
    def handle(self, *args, **options):
        for music_file in MusicFile.objects.all().select_related("music_directory"):
            if not os.path.isfile(music_file.full_path()):
                music_file.missing = True
                music_file.save()
                logger.warning(f"Missing file {music_file.full_path()}.")
        for dir in MusicDirectory.objects.all():
            for f_path in _walk_music_files(dir.local_path):
                data = _data_from_tags(f_path, dir.local_path)
                if (
                    data["artist"] and data["title"]
                ):  # absolute required minimum metadata !
                    MusicFile.objects.update_or_create(
                        music_directory=dir,
                        relative_path=data["relative_path"],
                        defaults={
                            "artist": data["artist"],
                            "title": data["title"],
                            "album": data["album"],
                            "genre": data["genre"],
                            "comment": data["comment"],
                            "duration": data["duration"],
                            "track_number": data["track"],
                            "created_at": data["creation"],
                            "modified_at": data["modification"],
                        },
                    )
                    logger.info(f"Added file {f_path}")
                else:
                    logger.warning(
                        f"Can't find artist and/or title for file {f_path}, will not add to index."
                    )
        fts_index_music_files()
        cache.clear()
