from django.db import models
from django.contrib.auth.models import User
import os
from django.utils.text import slugify
import mimetypes
from django.urls import reverse
from django.conf import settings


class MusicDirectory(models.Model):
    local_path = models.TextField(unique=True, null=False, blank=False)
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.local_path


class MusicFile(models.Model):
    music_directory = models.ForeignKey(MusicDirectory, on_delete=models.CASCADE)
    relative_path = models.TextField(unique=True, null=False, blank=False)
    artist = models.TextField(null=False, blank=False)
    title = models.TextField(null=False, blank=False)
    album = models.TextField(null=False)
    genre = models.TextField(null=False)
    comment = models.TextField(null=False)
    duration = models.IntegerField(null=False, default=0)
    track_number = models.IntegerField(null=False, default=0)
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)
    missing = models.BooleanField(default=False)

    def __str__(self) -> str:
        return f"{self.artist} - {self.title}"

    def full_path(self) -> str:
        return os.path.join(self.music_directory.local_path, self.relative_path)

    def download_url(self) -> str:
        if settings.DELEGATE_DOWNLOAD_PREFIX:
            return f"{settings.DELEGATE_DOWNLOAD_PREFIX}/{self.relative_path}"
        return reverse(
            "download_music_file",
            kwargs={
                "music_file_id": self.pk,
                "clean_file_name": self.clean_file_name(),
            },
        )

    def fts_data(self) -> str:
        return f"{self.artist} {self.title} {self.album} {self.genre} {self.comment}".strip()

    def file_extension(self) -> str:
        return os.path.splitext(self.relative_path)[1][1:]

    def mime_type(self) -> str:
        return mimetypes.guess_type(self.relative_path)[0]

    def clean_file_name(self) -> str:
        name_part = f"{self.artist} - {self.title}"[0:55]
        name_part = f"{name_part} - {self.pk}"
        return f"{slugify(name_part)}.{self.file_extension()}"

    class Meta:
        unique_together = (("music_directory", "relative_path"),)


class Tag(models.Model):
    owner = models.ForeignKey(User, on_delete=models.CASCADE, null=False, blank=False)
    label = models.CharField(max_length=255, null=False, blank=False)
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)
    is_playlist = models.BooleanField(default=False)

    def __str__(self):
        return self.label

    class Meta:
        unique_together = (("owner", "label"),)


class MusicFileTag(models.Model):
    tag = models.ForeignKey(Tag, on_delete=models.CASCADE, null=False, blank=False)
    music_file = models.ForeignKey(
        MusicFile, on_delete=models.CASCADE, null=False, blank=False
    )
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)
    rank = models.IntegerField(null=False, default=0)

    class Meta:
        unique_together = (("tag", "music_file"),)


class Preferences(models.Model):
    owner = models.OneToOneField(
        User, on_delete=models.CASCADE, null=False, blank=False
    )
    autofeed_queue = models.BooleanField(default=True)
    autofeed_similar = models.BooleanField(default=True)
    repeat_queue = models.BooleanField(default=True)
    current_position = models.IntegerField(null=False, default=0)
    access_token = models.TextField(null=True, unique=True)
