import json

from django.shortcuts import render
from django.http import HttpResponse, HttpRequest, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect
from reinhardt.models import MusicFile, Tag, MusicFileTag, Preferences
from .fileresponse import RangedFileResponse
from django.db.models.functions import Lower
from django.views.decorators.http import require_POST, require_safe
from django.contrib.auth.decorators import login_required
from django.db import connection, IntegrityError
from django.utils.translation import gettext as _
from django.contrib.auth import login
from .fts import fts_escape
import functools
import random
from .helpers import (
    add_music_file_to_tag,
    music_file_is_favorite,
    remove_music_file_from_tag,
    user_tags,
    add_tag_for_owner,
    remove_tag_for_owner,
    distinct_fuzzy,
    sanitize_tag_label,
    queue_list,
    user_preferences,
)


@functools.cache
def all_artists():
    artists = []
    for m in MusicFile.objects.values_list("artist").distinct():
        artists.append(m[0])
    return artists


@require_safe
@login_required
def index(request: HttpRequest) -> HttpResponse:
    music_files = MusicFile.objects.order_by("-created_at")[0:100]
    return render(request, "reinhardt/index.html", {"music_files": music_files})


@require_safe
@login_required
def artists(request: HttpRequest) -> HttpResponse:
    q = request.GET.get("q")
    if not q:
        letters = []
        sql = "select distinct(SUBSTRING(artist, 1, 1)) from reinhardt_musicfile order by artist;"
        with connection.cursor() as cursor:
            cursor.execute(sql)
            while row := cursor.fetchone():
                letters.append(row[0].upper())
        letters = list(set(letters))
        letters.sort()
        return render(request, "reinhardt/artists.html", {"letters": letters})
    else:
        artists = (
            MusicFile.objects.filter(artist__istartswith=q)
            .order_by(Lower("artist"))
            .values("artist")
            .distinct()
        )
        return render(
            request,
            "reinhardt/_artists_by_letter.html",
            {"artists": artists, "letter": q},
        )


@require_safe
@login_required
def genres(request: HttpRequest) -> HttpResponse:
    music_genres = (
        MusicFile.objects.exclude(genre="")
        .order_by(Lower("genre"))
        .values("genre")
        .distinct()
    )
    for genre in music_genres:
        genre["first_letter"] = genre["genre"].strip().lower()[0].strip()
    return render(request, "reinhardt/genres.html", {"music_genres": music_genres})


@require_safe
@login_required
def artist(request: HttpRequest) -> HttpResponse:
    q = request.GET.get("q")
    music_files = (
        MusicFile.objects.filter(artist=q).order_by("album", "track_number").distinct()
    )
    return render(
        request, "reinhardt/artist.html", {"music_files": music_files, "artist": q}
    )


@require_safe
@login_required
def genre(request: HttpRequest) -> HttpResponse:
    q = request.GET.get("q")
    music_files = (
        MusicFile.objects.filter(genre=q)
        .order_by("artist", "album", "track_number")
        .distinct()
    )
    return render(
        request, "reinhardt/genre.html", {"music_files": music_files, "genre": q}
    )


@require_safe
@login_required
def album(request: HttpRequest) -> HttpResponse:
    q = request.GET.get("q")
    music_files = MusicFile.objects.filter(album=q).order_by("track_number").distinct()
    return render(
        request, "reinhardt/album.html", {"music_files": music_files, "album": q}
    )


@require_safe
@login_required
def serve_music_file(
    request: HttpRequest, music_file_id: int, clean_file_name: str
) -> RangedFileResponse:
    music_file = get_object_or_404(MusicFile, pk=music_file_id)
    if clean_file_name != music_file.clean_file_name():
        return redirect(music_file.download_url())
    return RangedFileResponse(
        request,
        open(music_file.full_path(), "rb"),
        filename=music_file.clean_file_name(),
        content_type=music_file.mime_type(),
    )


@require_POST
@login_required
def add_to_tag(request: HttpRequest) -> HttpResponse:
    action = "added"
    object_label = ""
    tag_label = request.POST.get("tag")
    add_type = request.POST.get("add_type")
    q = request.POST.get("q")
    tag = request.POST.get("tag")
    if add_type == "music_file":
        object_label = "track"
        music_file = get_object_or_404(MusicFile, pk=q)
        add_music_file_to_tag(music_file, request.user, tag)
    if add_type == "album":
        object_label = "album"
        for music_file in MusicFile.objects.filter(album=q).order_by("track_number"):
            add_music_file_to_tag(music_file, request.user, tag)
    if add_type == "genre":
        object_label = "genre"
        for music_file in MusicFile.objects.filter(genre=q).order_by(
            "artist", "album", "track_number"
        ):
            add_music_file_to_tag(music_file, request.user, tag)
    if add_type == "artist":
        object_label = "artist"
        for music_file in MusicFile.objects.filter(artist=q).order_by(
            "album", "track_number"
        ):
            add_music_file_to_tag(music_file, request.user, tag)
    if add_type == "tag" and q != tag:
        object_label = "tag"
        for item in MusicFileTag.objects.filter(
            tag__label=q, tag__owner=request.user
        ).order_by(
            "music_file__genre",
            "music_file__artist",
            "music_file__album",
            "music_file__track_number",
        ):
            add_music_file_to_tag(item.music_file, request.user, tag)
    if add_type == "favorites" and tag != "favorites":
        object_label = "favorites"
        for item in MusicFileTag.objects.filter(
            tag__label="favorites", tag__owner=request.user
        ).order_by(
            "music_file__genre",
            "music_file__artist",
            "music_file__album",
            "music_file__track_number",
        ):
            add_music_file_to_tag(item.music_file, request.user, tag)
    return render(
        request,
        "reinhardt/_tag_message.html",
        {"object_label": object_label, "tag_label": tag_label, "action": action},
    )


@require_POST
@login_required
def toggle_tag(request: HttpRequest) -> HttpResponse:
    music_file_pk = request.POST.get("music_file_pk")
    has_tag = True if request.POST.get("has_tag") == "on" else False
    music_file = get_object_or_404(MusicFile, pk=music_file_pk)
    tag_instance = get_object_or_404(
        Tag.objects.exclude(label__in=["queue", "favorites"]),
        pk=request.POST.get("tag_pk"),
        owner=request.user,
    )
    if has_tag:
        action = "added"
        add_music_file_to_tag(music_file, request.user, tag_instance.label)
    else:
        action = "removed"
        remove_music_file_from_tag(music_file, request.user, tag_instance.label)
    return render(
        request,
        "reinhardt/_tag_message.html",
        {
            "object_label": "track",
            "tag_label": tag_instance.label,
            "action": action,
            "object_pk": music_file.pk,
        },
    )


@require_POST
@login_required
def remove_from_tag(request: HttpRequest) -> HttpResponse:
    action = "removed"
    object_label = ""
    object_pk = ""
    tag_label = request.POST.get("tag")
    remove_type = request.POST.get("remove_type")
    q = request.POST.get("q")
    tag = request.POST.get("tag")
    if remove_type == "music_file":
        object_label = "track"
        music_file = get_object_or_404(MusicFile, pk=q)
        object_pk = music_file.pk
        remove_music_file_from_tag(music_file, request.user, tag)
    return render(
        request,
        "reinhardt/_tag_message.html",
        {
            "object_label": object_label,
            "tag_label": tag_label,
            "action": action,
            "object_pk": object_pk,
        },
    )


@require_safe
@login_required
def favorites(request: HttpRequest) -> HttpResponse:
    music_files = MusicFile.objects.filter(
        musicfiletag__tag__label="favorites", musicfiletag__tag__owner=request.user
    ).order_by("genre", "-musicfiletag__rank")
    return render(request, "reinhardt/favorites.html", {"music_files": music_files})


@login_required
def tags(request: HttpRequest) -> HttpResponse:
    if request.method == "POST":
        action = request.POST.get("action")
        if action == "add_tag":
            add_tag_for_owner(request.POST.get("tag_label"), request.user)
        if action == "remove_tag":
            remove_tag_for_owner(request.POST.get("tag_pk"), request.user)
    return render(request, "reinhardt/tags.html", {"tags": user_tags(request.user)})


@login_required
def tag(request: HttpRequest) -> HttpResponse:
    queryset = Tag.objects.exclude(label__in=["queue", "favorites"])
    tag_instance = get_object_or_404(
        queryset, pk=request.GET.get("q"), owner=request.user
    )
    music_files = MusicFile.objects.filter(musicfiletag__tag=tag_instance).order_by()
    return render(
        request, "reinhardt/tag.html", {"tag": tag_instance, "music_files": music_files}
    )


@login_required
def music_file_tagger(request: HttpRequest) -> HttpResponse:
    music_file = get_object_or_404(MusicFile, pk=request.GET.get("q"))
    next_url = request.GET.get("next")
    assigned_tags = (
        Tag.objects.exclude(label__in=["queue", "favorites"])
        .filter(owner=request.user, musicfiletag__music_file=music_file)
        .values("pk")
        .distinct()
    )
    assigned_tags_pks = []
    for item in assigned_tags:
        assigned_tags_pks.append(item.get("pk"))

    return render(
        request,
        "reinhardt/music_file_tagger.html",
        {
            "tags": user_tags(request.user),
            "music_file": music_file,
            "next_url": next_url,
            "assigned_tags_pks": assigned_tags_pks,
        },
    )


@require_safe
@login_required
def search(request: HttpRequest) -> HttpResponse:
    sql = "select * from reinhardt_musicfile inner join music_index on music_index.id = reinhardt_musicfile.id where music_index.data match %s order by rank"
    sql_count = "select count(*) from reinhardt_musicfile inner join music_index on music_index.id = reinhardt_musicfile.id where music_index.data match %s order by rank"
    q = request.GET.get("sq", "").strip()
    found_genres = distinct_fuzzy("genre", q) if q else []
    found_artists = distinct_fuzzy("artist", q) if q else []
    found_albums = distinct_fuzzy("album", q) if q else []
    total_count = 0
    music_files = MusicFile.objects.none()
    if q:
        with connection.cursor() as cursor:
            cursor.execute(sql_count, (fts_escape(q),))
            total_count = cursor.fetchone()[0]
        music_files = MusicFile.objects.raw(sql, [fts_escape(q)])[0:1000]
    has_results = bool(
        total_count + len(found_artists) + len(found_albums) + len(found_genres)
    )
    return render(
        request,
        "reinhardt/search.html",
        {
            "sq": q,
            "q": q,
            "has_results": has_results,
            "music_files": music_files,
            "total_count": total_count,
            "found_genres": found_genres,
            "found_artists": found_artists,
            "found_albums": found_albums,
        },
    )


@require_safe
@login_required
def music_file_dropdown(request: HttpRequest) -> HttpResponse:
    music_file = get_object_or_404(MusicFile, pk=request.GET.get("q"))
    is_favorite = music_file_is_favorite(music_file, request.user)
    prefs = user_preferences(request.user)
    remove_from_tag_label = ""
    playing = request.GET.get("playing", "")
    tag_id = request.GET.get("tag", "")
    context = request.GET.get("context", "")
    if tag_id:
        tag_instance = get_object_or_404(Tag, pk=tag_id, owner=request.user)
        remove_from_tag_label = tag_instance.label
    return render(
        request,
        "reinhardt/_track_dropdown_menu_items.html",
        {
            "music_file": music_file,
            "is_favorite": is_favorite,
            "remove_from_tag": remove_from_tag_label,
            "playing": playing,
            "autofeed": prefs.autofeed_queue,
            "context": context,
        },
    )


@login_required
def rename_tag(request: HttpRequest, tag_id: int) -> HttpResponse:
    tag = get_object_or_404(Tag, pk=tag_id, owner=request.user)
    error_message = ""
    if request.method == "POST":
        label = sanitize_tag_label(request.POST.get("tag_label"))
        if not label:
            error_message = "tag name is empty"
        else:
            try:
                tag.label = label
                tag.save()
                return redirect("tags")
            except IntegrityError:
                error_message = "tag already exists"
    return render(
        request,
        "reinhardt/rename_tag.html",
        {"tag": tag, "error_message": error_message},
    )


@require_POST
@login_required
def next_song(request: HttpRequest) -> HttpResponse:
    music_file = None
    insert_music_file = None
    prefs = user_preferences(request.user)
    current_position = 0
    last_played_item = MusicFileTag.objects.filter(
        pk=prefs.current_position, tag__owner=request.user, tag__label="queue"
    ).first()
    if last_played_item:
        current_position = last_played_item.rank
    next_queue_item_id = request.POST.get("list_item")

    # This case is when user directly plays
    # an item from the queue page
    if next_queue_item_id:
        # no queue modification here:
        # just fetch item from queue, set current pos and play
        item = get_object_or_404(
            MusicFileTag,
            pk=next_queue_item_id,
            tag__owner=request.user,
            tag__label="queue",
        )
        current_position = item.rank
        music_file = item.music_file
        prefs.current_position = item.pk
        prefs.save()
    # All other cases
    else:
        # Did the user asked to queue a track and play it immediately ?
        next_track_id = request.POST.get("next_track")
        if next_track_id:
            insert_music_file = get_object_or_404(MusicFile, pk=next_track_id)
        queue = queue_list(request.user).filter(rank__gt=current_position)
        # This case when user clicks the "play now" (except on the queue page)
        if insert_music_file:
            music_file = insert_music_file
            item = add_music_file_to_tag(insert_music_file, request.user, "queue")
            current_position = item.rank
            prefs.current_position = item.pk
            prefs.save()
        else:
            # This case for normal queue list advance
            for item in queue:
                if item.rank > current_position:
                    music_file = item.music_file
                    current_position = item.rank
                    prefs.current_position = item.pk
                    prefs.save()
                    break
        # No next track... do we add (autofeed) one ?
        if not music_file and prefs.autofeed_queue:
            music_file = (
                MusicFile.objects.filter(artist=random.choice(all_artists()))
                .order_by("?")
                .first()
            )
            item = add_music_file_to_tag(music_file, request.user, "queue")
            current_position = item.rank
            prefs.current_position = item.pk
            prefs.save()
        # nothing to play anymore, reset position
        if not music_file:
            current_position = ""
            prefs.current_position = 0
            prefs.save()
    return render(
        request,
        "reinhardt/_now_playing.html",
        {"music_file": music_file, "current_position": current_position},
    )


@require_POST
@login_required
def activate_autofeed(request: HttpRequest) -> HttpResponse:
    prefs = user_preferences(request.user)
    prefs.autofeed_queue = True
    prefs.save()
    return HttpResponse(
        f'<div class="alert alert-light hideMeAfter5Seconds" id="alert_space">{_("autofeed activated")}</div>'
    )


@require_POST
@login_required
def deactivate_autofeed(request: HttpRequest) -> HttpResponse:
    prefs = user_preferences(request.user)
    prefs.autofeed_queue = False
    prefs.save()
    return HttpResponse(
        f'<div class="alert alert-light hideMeAfter5Seconds" id="alert_space">{_("autofeed deactivated")}</div>'
    )


@login_required
def queue(request: HttpRequest) -> HttpResponse:
    prefs = user_preferences(request.user)
    queue = queue_list(request.user)
    return render(
        request,
        "reinhardt/queue.html",
        {"queue": queue, "now_playing": prefs.current_position},
    )


@require_POST
@login_required
def clear_queue(request: HttpRequest) -> HttpResponse:
    prefs = user_preferences(request.user)
    queue = queue_list(request.user)
    queue.delete()
    return render(
        request,
        "reinhardt/queue.html",
        {"queue": queue, "now_playing": prefs.current_position},
    )


@require_POST
@login_required
def reorder_queue(request: HttpRequest) -> HttpResponse:
    try:
        queue = json.loads(request.POST.get("new_order", "[]"))
        for line in queue:
            item = MusicFileTag.objects.get(
                pk=line["item"], tag__owner=request.user, tag__label="queue"
            )
            item.rank = line["rank"]
            item.save()
        return HttpResponse("ok")
    except:
        return HttpResponseBadRequest()


def remove_from_queue(request: HttpRequest) -> HttpResponse:
    item = get_object_or_404(
        MusicFileTag,
        pk=request.POST.get("remove_item"),
        tag__owner=request.user,
        tag__label="queue",
    )
    item.delete()
    return HttpResponse(item.pk)


def invite(request: HttpRequest) -> HttpResponse:
    next_url = request.GET.get("next")
    if request.user.is_authenticated:
        return redirect(next_url or "index")
    token = request.GET.get("q")
    prefs = get_object_or_404(Preferences, access_token=token)
    login(request, prefs.owner)
    return redirect(next_url or "index")
